/*
 * Binary tree database functions for deleting a record.
 *
 * Copyright 2005 Andrew Wood, distributed under the Artistic License.
 */

#include "config.h"
#include "database.h"

#ifdef USING_INTERNAL_DB

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include "btreei.h"


/*
 * Reposition the given record in the binary tree, by finding an existing
 * record to link it to. Returns nonzero on error.
 */
int qdb_delete__reposition(qdb_t db, unsigned long offset, char *token)
{
	struct bt_record record;
	unsigned long offs, parent;
	qdb_datum key;
	int x;

	key.data = token;
	key.size = strlen(token);

	x = dbbt_find_token(db, key, &record, &offs, &parent);

	if (parent < 1)
		return 1;

	if (dbbt_read_record(db, parent, &record))
		return 1;

	if (x < 0) {
		record.lower = offset;
	} else {
		record.higher = offset;
	}

	if (dbbt_write_record(db, parent, &record))
		return 1;

	return 0;
}


/*
 * Delete the given key from the database. Returns nonzero on error.
 */
int qdb_delete(qdb_t db, qdb_datum key)
{
	struct bt_record record, recparent, reclower, rechigher;
	unsigned long offset, parent, nextfree;

	int x;

	if (db == NULL)
		return 1;

	if (key.size < 1)
		return 1;

	x = dbbt_find_token(db, key, &record, &offset, &parent);
	if (offset < 1)
		return 1;

	/*
	 * Get a copy of the lower and higher records, if any.
	 */
	if (record.lower > 0) {
		if (dbbt_read_record(db, record.lower, &reclower))
			return 1;
	}

	if (record.higher > 0) {
		if (dbbt_read_record(db, record.higher, &rechigher))
			return 1;
	}

	dbbt_sig_block();

	/*
	 * Remove the link to this record from its parent.
	 */
	if (parent > 0) {
		if (dbbt_read_record(db, parent, &recparent)) {
			dbbt_sig_unblock();
			return 1;
		}
		if (x < 0) {
			recparent.lower = 0;
		} else {
			recparent.higher = 0;
		}
		if (dbbt_write_record(db, parent, &recparent)) {
			dbbt_sig_unblock();
			return 1;
		}
	}


	/*
	 * Re-attach any children of this record to the database.
	 */
	if (record.lower > 0)
		qdb_delete__reposition(db, record.lower, reclower.token);
	if (record.higher > 0)
		qdb_delete__reposition(db, record.higher, rechigher.token);

	/*
	 * Now we add this record's offset to the head of the "free records"
	 * linked list.
	 */
	if (lseek(db->fd, sizeof(offset), SEEK_SET) == (off_t) - 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	nextfree = 0x80000000;
	if (dbbt_chunkread(&nextfree, sizeof(nextfree), 1, db->fd) < 0) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	if (lseek(db->fd, sizeof(offset), SEEK_SET) == (off_t) - 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	offset |= 0x80000000;
	if (dbbt_chunkwrite(&offset, sizeof(offset), 1, db->fd) < 0) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	offset &= 0x7FFFFFFF;

	if (lseek(db->fd, offset, SEEK_SET) == (off_t) - 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	if (dbbt_chunkwrite(&nextfree, sizeof(nextfree), 1, db->fd) < 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	dbbt_sig_unblock();

	return 0;
}

#endif				/* USING_INTERNAL_DB */

/* EOF */
