/*
 * Binary tree database functions for storing a record.
 *
 * Copyright 2005 Andrew Wood, distributed under the Artistic License.
 */

#include "config.h"
#include "database.h"

#ifdef USING_INTERNAL_DB

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>
#include "btreei.h"


/*
 * Store the given key with the given value into the database, replacing any
 * existing value for that key. Returns nonzero on error.
 */
int qdb_store(qdb_t db, qdb_datum key, qdb_datum val)
{
	struct bt_record record, head;
	unsigned long offset, parent, nextfree;
	int x;

	if (db == NULL)
		return 1;

	memset(&head, 0, BT_RECORD_SIZE);
	memset(&record, 0, BT_RECORD_SIZE);

	if (key.size >= BT_TOKEN_MAX)
		key.size = BT_TOKEN_MAX - 1;

	x = dbbt_find_token(db, key, &record, &offset, &parent);

	memcpy(record.token, key.data, key.size);
	record.token[key.size] = 0;
	record.data[0] = ((long *) val.data)[0];
	record.data[1] = ((long *) val.data)[1];

	dbbt_sig_block();

	/*
	 * Record exists - overwrite it.
	 */
	if (offset > 0) {
		if (dbbt_write_record(db, offset, &record)) {
			dbbt_sig_unblock();
			return 1;
		}
		dbbt_sig_unblock();
		return 0;
	}

	record.lower = 0;
	record.higher = 0;

	/*
	 * Database has just been created, so fill in the header and write
	 * this record as the first one.
	 */
	if (db->size <= sizeof(offset)) {

		if (lseek(db->fd, 0, SEEK_SET) == (off_t) - 1) {
			bt_lasterror = strerror(errno);
			dbbt_sig_unblock();
			return 1;
		}
		offset = 2 * sizeof(offset);
		if (dbbt_chunkwrite(&offset, sizeof(offset), 1, db->fd) <
		    1) {
			bt_lasterror = strerror(errno);
			dbbt_sig_unblock();
			return 1;
		}

		db->gotheadoffs = 0;

		offset = 0;
		if (dbbt_chunkwrite(&offset, sizeof(offset), 1, db->fd) <
		    1) {
			bt_lasterror = strerror(errno);
			dbbt_sig_unblock();
			return 1;
		}

		head.lower = (2 * sizeof(offset)) + BT_RECORD_SIZE;
		if (dbbt_chunkwrite(&head, BT_RECORD_SIZE, 1, db->fd) < 1) {
			bt_lasterror = strerror(errno);
			dbbt_sig_unblock();
			return 1;
		}

		if (dbbt_chunkwrite(&record, BT_RECORD_SIZE, 1, db->fd) <
		    1) {
			bt_lasterror = strerror(errno);
			dbbt_sig_unblock();
			return 1;
		}

		db->size = lseek(db->fd, 0, SEEK_CUR);
		dbbt_sig_unblock();
		return 0;
	}

	/*
	 * Get offset of next free space block.
	 */
	if (lseek(db->fd, sizeof(offset), SEEK_SET) == (off_t) - 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}
	if (dbbt_chunkread(&offset, sizeof(offset), 1, db->fd) < 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	offset &= 0x7FFFFFFF;

	/*
	 * If offset is 0 or we can't read from that offset, it's a new
	 * block at the end of the file, otherwise we take the next free
	 * offset from there and store it in the core free pointer, and then
	 * use that free offset.
	 */
	if (lseek(db->fd, offset, SEEK_SET) == (off_t) - 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}
	if ((offset == 0)
	    || (dbbt_chunkread(&nextfree, sizeof(nextfree), 1, db->fd) < 1)
	    ) {
		offset = lseek(db->fd, 0, SEEK_END);
		nextfree = 0x80000000;
	}
	if (lseek(db->fd, sizeof(offset), SEEK_SET) == (off_t) - 1) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}
	if (dbbt_chunkwrite(&nextfree, sizeof(nextfree), 1, db->fd) < 0) {
		bt_lasterror = strerror(errno);
		dbbt_sig_unblock();
		return 1;
	}

	if (dbbt_write_record(db, offset, &record)) {
		dbbt_sig_unblock();
		return 1;
	}

	/*
	 * Now attach the new record to its parent, if applicable.
	 */
	if (parent > 0) {
		if (dbbt_read_record(db, parent, &record)) {
			dbbt_sig_unblock();
			return 1;
		}
		if (x < 0) {
			record.lower = offset;
		} else {
			record.higher = offset;
		}
		if (dbbt_write_record(db, parent, &record)) {
			dbbt_sig_unblock();
			return 1;
		}
	}

	dbbt_sig_unblock();

	return 0;
}

#endif				/* USING_INTERNAL_DB */

/* EOF */
