/*
 * Rules which look for gibberish - words containing too many consonants or
 * vowels in a row.
 *
 * Copyright 2005 Andrew Wood, distributed under the Artistic License.
 */

#include "config.h"
#include "testi.h"
#include <string.h>

#define CONSONANTS "BCDFGHJKLMNPQRSTVWXZbcdfghjklmnpqrstvwxz" \
		   ""
#define VOWELS "AEIOUYaeiouy" \
	       "" \
	       "\0377"

/*
 * Return the next word in the message, or 0 if there are no more, if "get"
 * is 1, otherwise reset the position counter. The string returned is in an
 * internal static buffer.
 */
static char *spam_test_gibberish__getword(msg_t msg, int get)
{
	static char word[64];		 /* RATS: ignore (string bounded) */
	static long pos = 0;
	int wordpos;

	if (get == 0) {
		pos = 0;
		return 0;
	}

	word[0] = 0;
	wordpos = 0;

	/*
	 * Skip all white-space.
	 */
	while ((pos < msg->text_size)
	       && ((msg->textcontent[pos] == ' ')
		   || (msg->textcontent[pos] == '\t')
		   || (msg->textcontent[pos] == '\r')
		   || (msg->textcontent[pos] == '\n')
	       )
	    ) {
		pos++;
	}

	if (pos >= msg->text_size)
		return 0;

	while ((pos < msg->text_size)
	       && (wordpos < sizeof(word) - 1)
	       && (msg->textcontent[pos] != ' ')
	       && (msg->textcontent[pos] != '\t')
	       && (msg->textcontent[pos] != '\r')
	       && (msg->textcontent[pos] != '\n')
	    ) {
		word[wordpos] = msg->textcontent[pos];
		wordpos++;
		word[wordpos] = 0;
		pos++;
	}

	if (strlen(word) > 0)
		return word;

	return 0;
}


/*
 * Return nonzero if a sequence of "count" or more characters from the
 * string "accept" are found in the given word, so long as the word is
 * shorter than 60 characters (so we don't trip up on base64 encoded stuff).
 */
static int spam_test_gibberish__runon(char *word, char *accept, int count)
{
	int pos = 0;
	int n;

	if (strlen(word) >= 60)
		return 0;

	while (pos < strlen(word)) {
		pos += strcspn(word + pos, accept);
		if (pos >= strlen(word))
			return 0;
		n = strspn(word + pos, accept);
		pos += n;
		if (n >= count)
			return 1;
	}

	return 0;
}


/*
 * Add a token for every word found that contains more than 5 consonants
 * (not including "y") in a row.
 */
int spam_test_gibberish_consonants(opts_t opts, msg_t msg, spam_t spam)
{
	int nfound = 0;
	char *word;

	spam_test_gibberish__getword(msg, 0);

	while ((word = spam_test_gibberish__getword(msg, 1)) != 0) {
		if (spam_test_gibberish__runon(word, CONSONANTS, 5))
			nfound++;
	}

	if (nfound > 0)
		return nfound + 1;

	return 0;
}


/*
 * Add a token for every word found that contains more than 4 vowels
 * (including "y") in a row.
 */
int spam_test_gibberish_vowels(opts_t opts, msg_t msg, spam_t spam)
{
	int nfound = 0;
	char *word;

	spam_test_gibberish__getword(msg, 0);

	while ((word = spam_test_gibberish__getword(msg, 1)) != 0) {
		if (spam_test_gibberish__runon(word, VOWELS, 4))
			nfound++;
	}

	if (nfound > 0)
		return nfound + 1;

	return 0;
}


/*
 * Add a token for if the "From:" address contains more than 5 consonants
 * (not including "y") in a row.
 */
int spam_test_gibberish_from_consonants(opts_t opts, msg_t msg,
					spam_t spam)
{
	if (msg->sender == NULL)
		return 0;

	if (spam_test_gibberish__runon(msg->sender, CONSONANTS, 5))
		return 2;

	return 0;
}


/*
 * Add a token if the "From:" address contains more than 4 vowels (including
 * "y") in a row.
 */
int spam_test_gibberish_from_vowels(opts_t opts, msg_t msg, spam_t spam)
{
	if (msg->sender == NULL)
		return 0;

	if (spam_test_gibberish__runon(msg->sender, VOWELS, 4))
		return 2;

	return 0;
}

/* EOF */
